// -*- c-basic-offset: 2; related-file-name: "../include/click/packet.hh" -*-
/*
 * packet.{cc,hh} -- a packet structure. In the Linux kernel, a synonym for
 * `struct sk_buff'
 * Eddie Kohler, Robert Morris, Nickolai Zeldovich
 *
 * Copyright (c) 1999-2001 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, subject to the conditions
 * listed in the Click LICENSE file. These conditions include: you must
 * preserve this copyright notice, and you cannot mention the copyright
 * holders in advertising related to the Software without their permission.
 * The Software is provided WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED. This
 * notice is a summary of the Click LICENSE file; the license in that file is
 * legally binding.
 */

#include <click/config.h>

#include <click/packet.hh>
#include <click/glue.hh>
#ifdef CLICK_USERLEVEL
#include <unistd.h>
#endif
#include <assert.h>

#ifdef CLICK_LINUXMODULE	/* Linux kernel module */

Packet::Packet()
{
  static_assert(sizeof(Anno) <= sizeof(((struct sk_buff *)0)->cb));
  panic("Packet constructor");
}

Packet::~Packet()
{
  panic("Packet destructor");
}

WritablePacket *
Packet::make(uint32_t headroom, const unsigned char *data,
	     uint32_t len, uint32_t tailroom)
{
  int want = 1;
  if (struct sk_buff *skb = skbmgr_allocate_skbs(headroom, len + tailroom, &want)) {
    assert(want == 1);
    // packet comes back from skbmgr with headroom reserved
    __skb_put(skb, len);	// leave space for data
    if (data) memcpy(skb->data, data, len);
    skb->pkt_type = HOST | PACKET_CLEAN;
    WritablePacket *q = reinterpret_cast<WritablePacket *>(skb);
    q->clear_annotations();
    return q;
  } else
    return 0;
}

#else		/* User-space or BSD kernel module */

inline
Packet::Packet()
{
  _use_count = 1;
  _data_packet = 0;
  _head = _data = _tail = _end = 0;
  _destructor = 0;
#ifdef CLICK_BSDMODULE	/* BSD kernel module */
  _m = 0;
#endif
  clear_annotations();
}

Packet::~Packet()
{
  if (_data_packet)
    _data_packet->kill();
  else if (_head && _destructor)
    _destructor(_head, _end - _head);
  else {
#ifdef CLICK_USERLEVEL
    delete[] _head;
#endif

#ifdef CLICK_BSDMODULE
    m_freem(_m);
#endif
  }
  _head = _data = 0;
}

inline WritablePacket *
Packet::make(int, int, int)
{
  return static_cast<WritablePacket *>(new Packet(6, 6, 6));
}

#ifdef CLICK_USERLEVEL

WritablePacket *
Packet::make(unsigned char *data, uint32_t len, void (*destruct)(unsigned char *, size_t))
{
  WritablePacket *p = new WritablePacket;
  if (p) {
    p->_head = p->_data = data;
    p->_tail = p->_end = data + len;
    p->_destructor = destruct;
  }
  return p;
}

#endif

bool
Packet::alloc_data(uint32_t headroom, uint32_t len, uint32_t tailroom)
{
  uint32_t n = len + headroom + tailroom;
  if (n < MIN_BUFFER_LENGTH) {
    tailroom = MIN_BUFFER_LENGTH - len - headroom;
    n = MIN_BUFFER_LENGTH;
  }
  _destructor = 0;
#ifdef CLICK_BSDMODULE		/* BSD kernel module */
  if (n > MCLBYTES) {
    click_chatter("trying to allocate %d bytes: too many\n", n);
    return false;
  }
  MGETHDR(_m, M_WAIT, MT_DATA);
  if (!_m)
    return false;
  if (n > MHLEN) {
    MCLGET(_m, M_WAIT);
    if (!(_m->m_flags & M_EXT))
      return false;
  }
  _m->m_data += headroom;
  _m->m_len = len;
  _m->m_pkthdr.len = len;
  assimilate_mbuf();
#else			/* User-space */
  _head = new unsigned char[n];
  if (!_head)
    return false;
  _data = _head + headroom;
  _tail = _data + len;
  _end = _head + n;
#endif
  return true;
}

WritablePacket *
Packet::make(uint32_t headroom, const unsigned char *data, uint32_t len,
	     uint32_t tailroom)
{
  WritablePacket *p = new WritablePacket;
  if (!p)
    return 0;
  if (!p->alloc_data(headroom, len, tailroom)) {
    delete p;
    return 0;
  }
  if (data)
    memcpy(p->data(), data, len);
  return p;
}

#endif /* CLICK_LINUXMODULE */

//
// UNIQUEIFICATION
//

#ifdef CLICK_LINUXMODULE

Packet *
Packet::clone()
{
  struct sk_buff *nskb = skb_clone(skb(), GFP_ATOMIC);
  return reinterpret_cast<Packet *>(nskb);
}

WritablePacket *
Packet::expensive_uniqueify()
{
  // XXX would be better to reallocate data, but keep current header
  struct sk_buff *nskb = skb_copy(skb(), GFP_ATOMIC);
  if (nskb) {
    // all annotations, including IP header annotation, are copied,
    // but IP header will point to garbage if old header was 0
    if (!network_header())
      nskb->nh.raw = nskb->h.raw = 0;
  }
  kill();
  return reinterpret_cast<WritablePacket *>(nskb);
}

#else		/* User-level or BSD kernel module */

Packet *
Packet::clone()
{
    // timing: .31-.39 normal, .43-.55 two allocs, .55-.58 two memcpys
    Packet *p = Packet::make(6, 6, 6); // dummy arguments: no initialization
    if (!p)
	return 0;
    memcpy(p, this, sizeof(Packet));
    p->_use_count = 1;
    p->_data_packet = this;
    p->_destructor = 0;
#ifdef CLICK_BSDMODULE		/* BSD kernel module */
    p->_m = 0;
#endif
    // increment our reference count because of _data_packet reference
    _use_count++;
    return p;
}

WritablePacket *
Packet::expensive_uniqueify()
{
  // XXX would be better to reallocate data, but keep current header
  WritablePacket *p = Packet::make(6, 6, 6); // dummy arguments: no initialization
  if (p) {
    p->_use_count = 1;
    p->_data_packet = 0;
    if (p->alloc_data(headroom(), length(), tailroom())) {
      memcpy(p->_data, _data, _tail - _data);
      p->copy_annotations(this);
      if (_nh.raw) {
	p->_nh.raw = p->_data + network_header_offset();
	p->_h.raw = p->_data + transport_header_offset();
      } else {
	p->_nh.raw = 0;
	p->_h.raw = 0;
      }
    } else {
      delete p;
      p = 0;
    }
  }
  kill();
  return p;
}

#endif

#ifdef CLICK_BSDMODULE		/* BSD kernel module */

struct mbuf *
Packet::steal_m()
{
  struct Packet *p;
  struct mbuf *m2;

  p = uniqueify();
  m2 = p->m();

  /* Clear the mbuf from the packet: otherwise kill will MFREE it */
  p->_m = 0;
  p->kill();
  return m2;
}

#endif /* CLICK_BSDMODULE */

//
// EXPENSIVE_PUSH, EXPENSIVE_PUT
//

/*
 * Prepend some empty space before a packet.
 * May kill this packet and return a new one.
 */
WritablePacket *
Packet::expensive_push(uint32_t nbytes)
{
  static int chatter = 0;
  if (headroom() < nbytes && chatter < 5) {
    click_chatter("expensive Packet::push; have %d wanted %d",
                  headroom(), nbytes);
    chatter++;
  }
  WritablePacket *q = Packet::make((nbytes + 128) & ~3, buffer_data(), buffer_length(), 0);
  if (q) {
    // [N+128, H+L+T, 0 / H+L+T]
#ifdef CLICK_LINUXMODULE	/* Linux kernel module */
    sk_buff *skb = q->skb();
    skb->data -= nbytes - headroom();
    // [128+H, N+L+T, 0 / H+L+T]
    skb->tail -= tailroom();
    // [128+H, N+L, T / H+L+T]
    skb->len += nbytes - tailroom() - headroom();
    // [128+H, N+L, T / N+L]
#else				/* User-space and BSD kernel module */
    q->_data -= nbytes - headroom();
    // [128+H, N+L+T, 0]
    q->_tail -= tailroom();
    // [128+H, N+L, T]
#ifdef CLICK_BSDMODULE		/* BSD kernel module */
    q->m()->m_data -= nbytes - headroom();
    // [128+H, N+L+T, 0]
    q->m()->m_len -= tailroom();
    q->m()->m_pkthdr.len -= tailroom();
    // [128+H, N+L, T]
#endif
#endif
    q->copy_annotations(this);
    if (network_header())
      q->set_network_header(q->data() + network_header_offset() + nbytes, network_header_length());
  }
  kill();
  return q;
}

WritablePacket *
Packet::expensive_put(uint32_t nbytes)
{
  static int chatter = 0;
  if (tailroom() < nbytes && chatter < 5) {
    click_chatter("expensive Packet::put; have %d wanted %d",
                  tailroom(), nbytes);
    chatter++;
  }
  WritablePacket *q = Packet::make(0, buffer_data(), buffer_length(), nbytes + 128);
  if (q) {
#ifdef CLICK_LINUXMODULE
    sk_buff *skb = q->skb();
    skb->tail += nbytes - tailroom();
    skb->data += headroom();
    skb->len += nbytes - tailroom() - headroom();
#else				/* User-space and BSD module */
    q->_tail += nbytes - tailroom();
    q->_data += headroom();
#ifdef CLICK_BSDMODULE		/* BSD */
    q->m()->m_len += nbytes - tailroom();
    q->m()->m_pkthdr.len += nbytes - tailroom();
    q->m()->m_data += headroom();
#endif
#endif
    q->copy_annotations(this);
    if (network_header())
      q->set_network_header(q->data() + network_header_offset(), network_header_length());
  }
  kill();
  return q;
}
